#!python
# -*- coding: utf-8 -*-
__version__ = "$Revision: 1.20 $"
# $Source: /home/mechanoid/vhosts/mechanoid.home.lan/htdocs/content/ml-kalman.html/src/kalman/RCS/main.py,v $
#
#       OS : GNU/Linux 4.10.3-1-ARCH 
# COMPILER : Python 3.6.0
#
#   AUTHOR : Evgeny S. Borisov
# 
#    http://www.mechanoid.kiev.ua
#  e-mail : nn@mechanoid.kiev.ua
# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 

import sys
import numpy as np
import matplotlib.pyplot as plt

# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
def plot_2D_data(y,yn,yk,f):
    plt.clf()
    plt.grid()
    plt.plot( y[:,0], y[:,1], label='data',color='red' )
    plt.plot( yn[:,0], yn[:,1], label='data+noise',color='yellow' )
    plt.plot( yk[:,0], yk[:,1], label='kalman' )
    plt.legend(loc='upper left')
    plt.savefig(f, dpi=100)


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
# фильтр Калмана
#
def kalman_filter( z, 
                F = np.eye(2), # матрица перехода состояний системы (transitionMatrix)
                Q = np.eye(2)*3e-3, # ковариация шума (processNoiseCov)
                H = np.eye(2), # матрица наблюдения (measurement)
                R = np.eye(2)*3e-2 # ковариация шума наблюдения (measurementNoiseCov)
            ):

    n = z.shape[0] # количество элементов
    #m = z.shape[1] # размерность пространства

    x = z[0,:].transpose() # начальное значение предсказания состояния системы
    P = Q # ковариационная матрица для экстраполированного вектора состояния

    r = [ x ] # результат

    for i in range(1,n):

        # экстраполяция:
        x = F.dot( x ) # экстраполяция (предсказание) состояния системы
        P = F.dot(P).dot( F.transpose() ) + Q # ковариационная матрица для экстраполированного вектора состояния

        # коррекция:
        y = z[i] - H.dot(x) # вектор ошибки, отклонение фактического наблюдения от ожидаемого (экстраполированного) наблюдения 
        S = H.dot(P).dot(H.transpose()) + R # ковариационная матрица для вектора отклонения (вектора ошибки)
        K = P.dot(H.transpose()).dot(np.linalg.inv(S)) # матрица коэффициентов усиления Калмана 
        x = x + K.dot(y) # коррекция оценки вектора состояния 
        P = ( np.eye(z.shape[1]) - K.dot(H) ).dot(P) # ковариационная матрица оценки вектора состояния системы

        r.append( x.transpose() )

    return np.array(r)




# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
def main():
    n = 100 # длинна последовательности

    # 2 dimensions
    y0 = [ i**2 for i in np.linspace(-1.0,1.0,n) ] 
    y1 = [ np.sin(i)/(i+1.0) for i in np.linspace(0.0,10.0,n) ] 

    d  = np.array( list(zip(y0,y1)) )
    dn = d + (np.random.rand(n,2)-0.5)*np.array([2e-2,6e-2])
    r =  kalman_filter(dn)

    plot_2D_data(d,dn,r,"result/klm.png")



# - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - - 
if __name__ == "__main__":
    sys.exit(main())



